# Copyright (c) 2012, TamTam Research s.r.o. http://www.tamtamresearch.com
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met: 
# 
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution. 
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies, 
# either expressed or implied, of the FreeBSD Project.

import boto
import argparse
import sys
import plac
import csv
from itertools import izip, count
import dateutil.parser
import os.path
from StringIO import StringIO
import gzip
import time

def gunzip_data(data):
    '''Decomress data from string using gzip method.

    :type data: string
    :param data: string to be decompressed
    
    :rtype: string
    :return: decompressed string
    '''
    buf = StringIO(data)
    gzipper = gzip.GzipFile(fileobj=buf)
    data = gzipper.read(data)
    gzipper.close()
    #todo, check for possible memory leak, shall I close the buf?
    return data



@plac.annotations(
    profile_name=("""Name of boto profile to use for credentials""", "option"),
    aws_access_key_id=("Your AWS Access Key ID", "option"),
    aws_secret_access_key=("Your AWS Secret Access Key", "option"),
    bucket_name=("bucket name (default: %(default)s)", "positional"),
    csv_version_file=("name of CSV file with version_id", "positional", None, argparse.FileType("rb")),
    output_version_id_names=("Resulting file names shall use version_id to become distinguished (default is to use timestamp of file creation)", "flag"),
    no_decompression=("Keeps the files as they come, do not decompress, if they come compressed", "flag")
)
def main(bucket_name, csv_version_file, output_version_id_names = False, no_decompression=False, profile_name=None,  aws_access_key_id=None,
          aws_secret_access_key=None):
    """Fetch file versions as listed in provided csv file
    
    Typical csv file (as by default produced by s3lsvers) is:
    
        my/versioned/feed.xml;OrUr6XO8KSKEHbd8mQ.MloGcGlsh7Sir;191345;2012-05-23T20:45:10.000Z;39
        my/versioned/feed.xml;xhkVOy.dJfjSfUwse8tsieqjDicp0owq;192790;2012-05-23T20:44:31.000Z;62
        my/versioned/feed.xml;oKneK.N2wS8pW8.EmLqjldYlgcFwxN3V;193912;2012-05-23T20:43:29.000Z;58

    and has columns:
    :key_name: name of the feed (not containing the bucket name itself)
    :version_id: string, identifying unique version. Any following columns can contain anything.
    :size: size in bytes. This column is not used and can be missing.
    :last_modified: date, when the version was posted. This column is not used and can be missing.
    
    Typical use (assuming, above csv file is available under name verlist.csv)::
    
        $ %(prog)s yourbucketname verlist.csv
    
    What will create following files in current directory:
    
    * my/versioned/feed.xml.2012-05-23T20_45_10.xml
    * my/versioned/feed.xml.2012-05-23T20_44_31.xml
    * my/versioned/feed.xml.2012-05-23T20_43_29.xml
    
    Even though these files are gzipped on server, they will be decompressed on local disk.
    
    """
    try:
        conn = boto.connect_s3(profile_name=profile_name, aws_access_key_id= aws_access_key_id, aws_secret_access_key= aws_secret_access_key)
        bucket = conn.get_bucket(bucket_name)
    except:
        print "Problem to connect to bucket or buckeet missing. Check your credentials (e.g. using BOTO_CONFIG)"
        return
    try:
        for i, row in izip(count(), csv.reader(csv_version_file, delimiter = ";")):
            key_name, version_id = row[:2]
            print "{i:2d}: key_name: {key_name}, version_id: {version_id}".format(**locals())
            key = bucket.get_key(key_name, version_id = version_id)
            last_modified = dateutil.parser.parse(key.last_modified)
            fname, ext = os.path.splitext(os.path.basename(key_name))
            if output_version_id_names:
                verpart = key.version_id
            else:
                verpart = "{last_modified:%Y-%m-%dT%H_%M_%SZ}".format(last_modified = last_modified)
            fname = "{fname}.{verpart}{ext}".format(**locals())
            with open(fname, "wb") as f:
                if (key.content_encoding in ("deflate", "gzip")) and (not no_decompression):
                    f.write(gunzip_data(key.get_contents_as_string()))
                else:
                    key.get_contents_to_file(f)
            #try to set modification time to the time, version was created in bucket.
            tm = time.mktime(last_modified.timetuple())
            os.utime(fname, (tm, tm))
            print "resulting file name: {fname}".format(fname = fname)
    except KeyboardInterrupt:
        print "...terminated."
        
    return

def placer():
    plac.call(main)
    
if __name__ == "__main__":
    plac.call(main)



from __future__ import print_function
import plac
import boto
import sys
import calendar
import datetime


def error(*objs):
    print("ERROR", *objs, end='\n', file=sys.stderr)


@plac.annotations(
    profile_name=("""Name of boto profile to use for credentials""", "option"),
    aws_access_key_id=("Your AWS Access Key ID", "option"),
    aws_secret_access_key=("Your AWS Secret Access Key", "option"),
    validate_bucket=("Make sure, the bucket really exists", "flag"),
    validate_key=("Make sure, the key really exists", "flag"),
    expire_dt=("ISO formatted time of expiration, full seconds,"
               " 'Z' is obligatory, e.g. '2014-02-14T21:47:16Z'"),
    bucket_name="name of bucket", key_names="key names to generate tmpurl for")
def main(
        expire_dt,
        bucket_name,
        profile_name=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        validate_bucket=False,
        validate_key=False,
        *key_names):
    """Generate temporary url for accessing content of AWS S3 key.

    Temporary url includes expiration time, after which it rejects serving the
    content.

    Urls are printed one per line to stdout.

    For missing key names empty line is printed and error goes to stderr.

    If the bucket is versioned, tmp url will serve the latest version
    at the moment of request (version_id is not part of generated url).

    By default, bucket and key name existnence is not verified.
    """
    expire_dt = datetime.datetime.strptime(expire_dt, "%Y-%m-%dT%H:%M:%SZ")
    expire = calendar.timegm(expire_dt.timetuple())
    try:
        con = boto.connect_s3(profile_name=profile_name)
    except Exception as e:
        error("Unable to connect to AWS S3, check your credentials", e)
        return
    try:
        validate = validate_bucket or validate_key
        bucket = con.get_bucket(bucket_name, validate=validate)
    except boto.exception.S3ResponseError as e:
        error("Error: Bucket not found: ", bucket_name)
        return
    for key_name in key_names:
        if validate_key:
            key = bucket.get_key(key_name)
            if key is None:
                error("Error: missing key: ", key_name)
                print("")
                continue
        # get a key object without version_id
        # (otherwise we get url to specific version)
        key = boto.s3.key.Key(bucket, key_name)
        print(key.generate_url(expires_in=expire, expires_in_absolute=True))


def placer():
    try:
        plac.call(main)
    except Exception as e:
        error(e)

if __name__ == "__main__":
    placer()



# Copyright (c) 2011, Jan Vlcinsky
# Copyright (c) 2012, TamTam Research s.r.o. http://www.tamtamresearch.com
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.

from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.4.3'

install_requires = [
    "setuptools",
    "PyYAML",
    "argparse",
    "boto>=2.25.0",
    "plac",
    "python-dateutil"
]

setup(name='ttr.aws.utils.s3',
      version=version,
      description=("Some command lines utilities to interact with files"
                   " stored in AWS S3 incl. versioned ones."),
      long_description=README + '\n\n' + NEWS,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Console",
          "License :: OSI Approved :: BSD License",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Topic :: Communications :: File Sharing",
          "Topic :: System :: Archiving",
          "Topic :: System :: Archiving :: Backup",
          "Topic :: System :: Recovery Tools",
          "Topic :: Utilities",
      ],
      keywords='aws s3 boto',
      author='Jan Vlcinsky',
      author_email='jan.vlcinsky@tamtamresearch.com',
      url='https://bitbucket.org/tamtamresearch/ttr.aws.utils.s3',
      license='BSD',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['ttr', 'ttr.aws', 'ttr.aws.utils'],
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      entry_points={
          'console_scripts':
          ['s3lsvers=ttr.aws.utils.s3.lsvers:placer',
             's3getvers=ttr.aws.utils.s3.getvers:placer',
             's3tmpgen=ttr.aws.utils.s3.tmpgen:placer'
           ]
          }
      )



# Copyright (c) 2011, Jan Vlcinsky
# Copyright (c) 2012, TamTam Research s.r.o. http://www.tamtamresearch.com
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.

import boto
import argparse
import sys
import datetime
import itertools
import dateutil.parser
import calendar
import plac
import os
import json
from csv import DictWriter
from ConfigParser import SafeConfigParser

from pkg_resources import resource_string
templ = resource_string(__name__, 'chart.html')


def load_alias(bucket_key):
    config = SafeConfigParser()
    rcfname = ".s3lsversrc"
    flocs = [os.curdir, os.path.expanduser("~"), "/etc/s3lsvers"]
    config.read([os.path.join(loc, rcfname) for loc in flocs])
    return config.get("aliases", bucket_key)


def get_json_record(dtime, size, age):
    tt = datetime.datetime.timetuple(dtime)
    epoch = calendar.timegm(tt) * 1000
    return [epoch, size, age]


@plac.annotations(
    profile_name=("""Name of boto profile to use for credentials""", "option"),
    aws_access_key_id=("Your AWS Access Key ID", "option"),
    aws_secret_access_key=("Your AWS Secret Access Key", "option"),
    from_time=(
        """Modification time of oldest expected version as ISO 8601 format.
        Can be truncated. (default: goes to the oldest version)""",
        "option", "from"),
    to_time=(
        """Modification time of youngest expected version as ISO 8601 format.
        Can be truncated. (default: goes to the latest version)""",
        "option", "to"),
    list_file=(
        """Name of file, where is result written in csv format.
        If set, the file is always overwritten.""",
        "option", None, argparse.FileType("wb")),
    html_file=(
        """Name of file, where is result written in html format (as a chart).
        If set, the file is always overwritten.""",
        "option", None, argparse.FileType("wb")),
    version_id=('''Optional version-id.
        If specified, listing does not start from the freshest version,
        but starts searching from given VERSION_ID
        and continues searching older and older versions.
        This could speed up listng in case, you need rather older files
        and you know VERSION_ID which came somehow later
        then is the time scope you are going to list.''',
                "option"),
    bucket_key=("bucket_name/key_name for the key to list, or key alias defined in .s3lsvers file", "positional")
)
def main(bucket_key, from_time=None, to_time=None,
         list_file=None, html_file=None, version_id=None, profile_name=None,  aws_access_key_id=None,
          aws_secret_access_key=None):
    """Lists all versions of given key, possibly filtering by from - to range
    for version last_modified time.
    Allows to put the listing into csv file and or into html chart.

        Listing shows:
          key_name
            "file name". Can repeat if the file has more versions

          version_id
            unique identifier for given version on given bucket.  Has form of
            string and not a number. identifiers are "random", do not expect
            that they are sorten alphabetically.

          size
            size of file in bytes

          last_modified
            ISO 8601 formated time of file modification,
            e.g. `2011-06-22T03:05:09.000Z`

          age
            difference between last_modified or given version
            and preceding version. It is sort of current
            update interval for that version.

        Sample use:
        Lists to the screen all versions of file keyname in the
        bucketname bucket::

            $ s3lsvers bucketname/keyname

        Lists all versions younger then given time (from given time till now)::

            $ s3lsvers -from 2011-07-19T12:00:00 bucketname/keyname

        Lists all versions older then given time
        (from very first version till given date)::

            $ s3lsvers -to 2011-07-19T12:00:00 bucketname/keyname

        Lists all versions in period betwen from and to time::

            $ s3lsvers -from 2010-01-01 -to 2011-07-19T12:00:00 \
            bucketname/keyname

        Lists all versions and writes them into csv file named versions.csv::

            $ s3lsvers -list-file versions.csv bucketname/keyname

        Lists all versions and writes them into html chart file
        named chart.html::

            $ s3lsvers -html-file chart.html bucketname/keyname

        Prints to screen, writes to csv, creates html chart and this all
        for versions in given time period.::

            $ s3lsvers -from 2010-01-01 -to 2011-07-19T12:00:00 \
            -list-file versions.csv -html-file chart.html bucketname/keyname

        Using bucket/key_name aliases in .s3lsvers file

        Instead of using long bucket and key names on command line, you may define aliases.

        Aliases are specified in file .s3lsvers, which may be located in currect directory, home directory or /etc/s3lsvers"

        Content of .s3lsvers files may look like this::

            #.s3lsversrc - definition of some preconfigured bucket/key values
            [DEFAULT]
            pl-base: pl-base.dp.tamtamresearch.com
            cz-base: cz-base.dp.tamtamresearch.com
            sk-base: sk-base.dp.tamtamresearch.com

            #values left to ":" must not contain "/" to prevent confusion with real bucket names
            [aliases]
            plcsr: %(pl-base)s/region/pl/ConsumerServiceReady.xml
            pldfs: %(pl-base)s/region/pl/DataFusionService.xml
            czcsr: %(cz-base)s/region/cz/ConsumerServiceReady.xml
            czdfs: %(cz-base)s/region/cz/DataFusionService.xml
            skcsr: %(sk-base)s/region/sk/ConsumerServiceReady.xml
            skdfs: %(sk-base)s/region/sk/DataFusionService.xml
            skes: %(sk-base)s/region/sk/EventService.xml
            sksr: %(sk-base)s/region/sk/SummaryReports.xml

        The format follows SafeConfigParser rules: http://docs.python.org/2/library/configparser.html#safeconfigparser-objects

    """

    if ("/" not in bucket_key):
        bucket_key = load_alias(bucket_key)
    assert "/" in bucket_key, "bucket_key must be in form 'bucket/key_name'"
    bucket_name, key_name = bucket_key.split("/", 1)
    cmdname = os.path.basename(sys.argv[0])
    if (from_time and to_time) and (from_time > to_time):
        msg = "-from ({from_time}) must be smaller then -to ({to_time})"
        print msg.format(**locals())
        return

    conn = boto.connect_s3(profile_name=profile_name, aws_access_key_id= aws_access_key_id, aws_secret_access_key= aws_secret_access_key)

    bucket = conn.get_bucket(bucket_name)
    try:
        if version_id is None:  # todo
            raise NameError
        versions = bucket.list_versions(
            key_name, key_marker=key_name, version_id_marker=version_id)
    except NameError:
        versions = bucket.list_versions(key_name)
    try:
        json_data = []
        if list_file:
            fields = ["key_name", "version_id", "size", "last_modified", "age"]
            csvwriter = DictWriter(
                list_file, fields, delimiter=";", extrasaction="ignore")
        for ver in vergen(versions, from_time, to_time, working, 600):
            print "Date: {last_modified} Size: {size} Age: {age}".format(**ver)
            if list_file:
                try:
                    csvwriter.writerow(ver)
                except Exception as e:
                    print e
                    raise
            if html_file:
                json_data.append(get_json_record(ver[
                                 "atime"], ver["size"], ver["age"]))

        if html_file:
            title = "{bucket_name}:{key_name}".format(**locals())
            json_data = json.dumps(json_data, separators=(",", ":"))
            res = render_html(json_data, title=title)
            html_file.write(res)
    except KeyboardInterrupt:
        print "...cancelled."
        pass
    finally:
        try:
            list_file.close()
        except:
            pass
        try:
            html_file.close()
        except:
            pass
    return


def working(ver):
    print "version num: {i}, modified: {last_modified}".format(**ver)


def vergen(versions, from_time, to_time, callback=None, callback_after=600):
    """return data
    atime, ka.size. age
    ka.name, ka.version_id, ka.size, ka.last_modified, age

    """
    vers_a, vers_b = itertools.tee(versions)
    vers_b.next()

    def buildver():
        """create dict, describing one version of key"""
        return {"i": i, "atime": atime, "btime": btime, "age": age,
                "key_name": ka.name, "size": ka.size,
                "last_modified": ka.last_modified, "version_id": ka.version_id
                }
    for i, ka, kb in itertools.izip(itertools.count(), vers_a, vers_b):
        last_modified = ka.last_modified
        atime = dateutil.parser.parse(last_modified)
        btime = dateutil.parser.parse(kb.last_modified)
        age = (atime - btime).seconds
        if from_time and (last_modified < from_time):
            break
        elif to_time and (last_modified > to_time):
            if i % callback_after == 0:
                callback(buildver())
            continue
        yield buildver()


def render_html(json_data, title,
                subtitle="(gzipped) feed size [bytes] and update intervals[seconds]",
                set_1_name="Feed size", set_2_name="Update iterval", ):
    return templ.format(**locals())


def placer():
    try:
        plac.call(main)
    except Exception as e:
        print e

if __name__ == "__main__":
    placer()
